@@ -73,7 +73,7 @@ public:
 
 	virtual Common::SeekableReadStream *createReadStream();
 	virtual Common::WriteStream *createWriteStream();
-	virtual bool create(bool isDirectory);
+	virtual bool create(bool isDirectoryFlag);
 
 private:
 	/**
