@@ -8,12 +8,12 @@
  * modify it under the terms of the GNU General Public License
  * as published by the Free Software Foundation; either version 2
  * of the License, or (at your option) any later version.
-
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
-
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
@@ -39,14 +39,14 @@ bool Sprite::contains(const Common::Point &pos) const {
 		return false;
 	if (adjustedPos.y < 0 || adjustedPos.y >= _surface.h)
 		return false;
-	byte *pixels = (byte *)_surface.pixels;
+	const byte *pixels = (const byte *)_surface.getPixels();
 	return (pixels[(_surface.h - adjustedPos.y - 1) * _surface.w + adjustedPos.x] != 0);
 }
 
 enum {
-	kAnimOpEvent = 1,
-	kAnimOpPlayWave = 2,
-	kAnimOpPlayAnim = 3,
+	kAnimOpEvent = 1, 
+	kAnimOpPlayWave = 2, 
+	kAnimOpPlayAnim = 3, 
 	kAnimOpDrawSprite = 4
 };
 
@@ -57,6 +57,7 @@ Animation::Animation(Common::SeekableReadStream *stream, uint16 id, Common::Poin
 
 	// probably total size?
 	uint32 unknown = _stream->readUint32LE();
+	_size = unknown;
 
 	debug(8, "anim: size %d, state %08x, unknown %08x", size, _state, unknown);
 
@@ -82,6 +83,55 @@ void Animation::seekToCurrPos() {
 	_stream->seek(_offset, SEEK_SET);
 }
 
+void ComposerEngine::loadAnimation(Animation *&anim, uint16 animId, int16 x, int16 y, int16 eventParam, int32 size) {
+	Common::SeekableReadStream *stream = NULL;
+	Pipe *newPipe = NULL;
+
+	// First, check the existing pipes.
+	for (Common::List<Pipe *>::iterator j = _pipes.begin(); j != _pipes.end(); j++) {
+		Pipe *pipe = *j;
+		if (!pipe->hasResource(ID_ANIM, animId))
+			continue;
+		stream = pipe->getResource(ID_ANIM, animId, false);
+
+		// When loading from savegame, make sure we have the correct stream
+		if ((!size) || (stream->size() >= size)) break;
+		stream = NULL;
+	}
+
+	// If we didn't find it, try the libraries.
+	if (!stream) {
+		if (!hasResource(ID_ANIM, animId)) {
+			warning("ignoring attempt to play invalid anim %d", animId);
+			return;
+		}
+		Common::List<Library>::iterator j;
+		for (j = _libraries.begin(); j != _libraries.end(); j++) {
+			stream = j->_archive->getResource(ID_ANIM, animId);
+
+			// When loading from savegame, make sure we have the correct stream
+			if ((!size) || (stream->size() >= size)) break;
+			stream = NULL;
+		}
+
+		uint32 type = j->_archive->getResourceFlags(ID_ANIM, animId);
+
+		// If the resource is a pipe itself, then load the pipe
+		// and then fish the requested animation out of it.
+		if (type != 1) {
+			_pipeStreams.push_back(stream);
+			newPipe = new Pipe(stream, animId);
+			_pipes.push_front(newPipe);
+			newPipe->nextFrame();
+			stream = newPipe->getResource(ID_ANIM, animId, false);
+		}
+	}
+
+	anim = new Animation(stream, animId, Common::Point(x, y), eventParam);
+	if (newPipe)
+		newPipe->_anim = anim;
+}
+
 void ComposerEngine::playAnimation(uint16 animId, int16 x, int16 y, int16 eventParam) {
 	// First, we check if this animation is already playing,
 	// and if it is, we sabotage that running one first.
@@ -93,49 +143,10 @@ void ComposerEngine::playAnimation(uint16 animId, int16 x, int16 y, int16 eventP
 		stopAnimation(*i);
 	}
 
-	Common::SeekableReadStream *stream = NULL;
-	Pipe *newPipe = NULL;
-
-	// First, check the existing pipes.
-	for (Common::List<Pipe *>::iterator j = _pipes.begin(); j != _pipes.end(); j++) {
-		Pipe *pipe = *j;
-		if (!pipe->hasResource(ID_ANIM, animId))
-			continue;
-		stream = pipe->getResource(ID_ANIM, animId, false);
-		break;
-	}
-
-	// If we didn't find it, try the libraries.
-	if (!stream) {
-		if (!hasResource(ID_ANIM, animId)) {
-			warning("ignoring attempt to play invalid anim %d", animId);
-			return;
-		}
-		stream = getResource(ID_ANIM, animId);
-
-		uint32 type = 0;
-		for (Common::List<Library>::iterator i = _libraries.begin(); i != _libraries.end(); i++)
-			if (i->_archive->hasResource(ID_ANIM, animId)) {
-				type = i->_archive->getResourceFlags(ID_ANIM, animId);
-				break;
-			}
-
-		// If the resource is a pipe itself, then load the pipe
-		// and then fish the requested animation out of it.
-		if (type != 1) {
-			_pipeStreams.push_back(stream);
-			newPipe = new Pipe(stream);
-			_pipes.push_front(newPipe);
-			newPipe->nextFrame();
-			stream = newPipe->getResource(ID_ANIM, animId, false);
-		}
-	}
-
-	Animation *anim = new Animation(stream, animId, Common::Point(x, y), eventParam);
+	Animation *anim = NULL;
+	loadAnimation(anim, animId, x, y, eventParam);
 	_anims.push_back(anim);
 	runEvent(kEventAnimStarted, animId, eventParam, 0);
-	if (newPipe)
-		newPipe->_anim = anim;
 }
 
 void ComposerEngine::stopAnimation(Animation *anim, bool localOnly, bool pipesOnly) {
@@ -376,7 +387,7 @@ void ComposerEngine::playPipe(uint16 id) {
 	}
 
 	Common::SeekableReadStream *stream = getResource(ID_PIPE, id);
-	OldPipe *pipe = new OldPipe(stream);
+	OldPipe *pipe = new OldPipe(stream, id);
 	_pipes.push_front(pipe);
 	//pipe->nextFrame();
 
@@ -541,7 +552,7 @@ void ComposerEngine::redraw() {
 
 	for (uint i = 0; i < _dirtyRects.size(); i++) {
 		const Common::Rect &rect = _dirtyRects[i];
-		byte *pixels = (byte *)_screen.pixels + (rect.top * _screen.pitch) + rect.left;
+		byte *pixels = (byte *)_screen.getBasePtr(rect.left, rect.top);
 		_system->copyRectToScreen(pixels, _screen.pitch, rect.left, rect.top, rect.width(), rect.height());
 	}
 	_system->updateScreen();
@@ -794,7 +805,7 @@ bool ComposerEngine::initSprite(Sprite &sprite) {
 
 	if (width > 0 && height > 0) {
 		sprite._surface.create(width, height, Graphics::PixelFormat::createFormatCLUT8());
-		decompressBitmap(type, stream, (byte *)sprite._surface.pixels, size, width, height);
+		decompressBitmap(type, stream, (byte *)sprite._surface.getPixels(), size, width, height);
 	} else {
 		// there are some sprites (e.g. a -998x-998 one in Gregory's title screen)
 		// which have an invalid size, but the original engine doesn't notice for
@@ -814,13 +825,13 @@ void ComposerEngine::drawSprite(const Sprite &sprite) {
 	int y = sprite._pos.y;
 
 	// incoming data is BMP-style (bottom-up), so flip it
-	byte *pixels = (byte *)_screen.pixels;
+	byte *pixels = (byte *)_screen.getPixels();
 	for (int j = 0; j < sprite._surface.h; j++) {
 		if (j + y < 0)
 			continue;
 		if (j + y >= _screen.h)
 			break;
-		byte *in = (byte *)sprite._surface.pixels + (sprite._surface.h - j - 1) * sprite._surface.w;
+		const byte *in = (const byte *)sprite._surface.getBasePtr(0, sprite._surface.h - j - 1);
 		byte *out = pixels + ((j + y) * _screen.w) + x;
 		for (int i = 0; i < sprite._surface.w; i++)
 			if ((x + i >= 0) && (x + i < _screen.w) && in[i])
