@@ -39,9 +39,8 @@ namespace Fullpipe {
 
 struct Bat {
 	StaticANIObject *ani;
-	int field_4;
 	double power;
-	double field_10;
+	double angle;
 	double currX;
 	double currY;
 	double powerCos;
@@ -54,15 +53,7 @@ void scene27_initScene(Scene *sc) {
 	g_vars->scene27_maid = sc->getStaticANIObject1ById(ANI_MAID, -1);
 	g_vars->scene27_batHandler = sc->getStaticANIObject1ById(ANI_BITAHANDLER, -1);
 
-	g_vars->scene27_balls.numBalls = 0;
-	g_vars->scene27_balls.pTail = 0;
-	g_vars->scene27_balls.field_8 = 0;
-	g_vars->scene27_balls.pHead = 0;
-	g_vars->scene27_balls.cPlexLen = 10;
-
-	free(g_vars->scene27_balls.cPlex);
-	g_vars->scene27_balls.cPlex = 0;
-
+	g_vars->scene27_balls.clear();
 	g_vars->scene27_bats.clear();
 	g_vars->scene27_var07.clear();
 
@@ -72,43 +63,7 @@ void scene27_initScene(Scene *sc) {
 	for (int i = 0; i < 4; i++) {
 		StaticANIObject *newbat = new StaticANIObject(g_vars->scene27_bat);
 
-		Ball *runPtr = g_vars->scene27_balls.pTail;
-		Ball *lastP = g_vars->scene27_balls.field_8;
-
-		if (!g_vars->scene27_balls.pTail) {
-			g_vars->scene27_balls.cPlex = (byte *)calloc(g_vars->scene27_balls.cPlexLen, sizeof(Ball));
-
-			byte *p1 = g_vars->scene27_balls.cPlex + (g_vars->scene27_balls.cPlexLen - 1) * sizeof(Ball);
-
-			if (g_vars->scene27_balls.cPlexLen - 1 < 0) {
-				runPtr = g_vars->scene27_balls.pTail;
-			} else {
-				runPtr = g_vars->scene27_balls.pTail;
-
-				for (int j = 0; j < g_vars->scene27_balls.cPlexLen; j++) {
-					((Ball *)p1)->p1 = runPtr;
-					runPtr = (Ball *)p1;
-
-					p1 -= sizeof(Ball);
-				}
-
-				g_vars->scene27_balls.pTail = runPtr;
-			}
-		}
-
-		g_vars->scene27_balls.pTail = runPtr->p0;
-		runPtr->p1 = lastP;
-		runPtr->p0 = 0;
-		runPtr->ani = newbat;
-
-		g_vars->scene27_balls.numBalls++;
-
-		if (g_vars->scene27_balls.field_8)
-			g_vars->scene27_balls.field_8->p0 = runPtr;
-		else
-			g_vars->scene27_balls.pHead = runPtr;
-
-		g_vars->scene27_balls.field_8 = runPtr;
+		g_vars->scene27_balls.push_back(newbat);
 
 		sc->addStaticANIObject(newbat, 1);
 	}
@@ -159,7 +114,7 @@ void sceneHandler27_driverGiveVent() {
 
 	g_vars->scene27_driverHasVent = false;
 
-	getCurrSceneSc2MotionController()->setEnabled();
+	getCurrSceneSc2MotionController()->activate();
 	getGameLoaderInteractionController()->enableFlag24();
 
 	g_fp->_behaviorManager->setFlagByStaticAniObject(g_fp->_aniMan, 1);
@@ -169,7 +124,7 @@ void sceneHandler27_winArcade() {
 	if (g_fp->getObjectState(sO_Driver) == g_fp->getObjectEnumState(sO_Driver, sO_WithSteering)) {
 		g_vars->scene27_dudeIsAiming = false;
 
-		g_fp->_aniMan->_callback2 = 0;
+		g_fp->_aniMan->_callback2 = 0; // Really NULL
 		g_fp->_aniMan->changeStatics2(ST_MAN_RIGHT);
 
 		sceneHandler27_driverGiveVent();
@@ -191,7 +146,7 @@ void sceneHandler27_showNextBat() {
 	if (g_vars->scene27_bat) {
 		MessageQueue *mq = new MessageQueue(g_fp->_currentScene->getMessageQueueById(QU_SC27_SHOWBET), 0, 1);
 
-		mq->replaceKeyCode(-1, g_vars->scene27_bat->_okeyCode);
+		mq->setParamInt(-1, g_vars->scene27_bat->_odelay);
 		mq->chain(0);
 	}
 
@@ -226,7 +181,7 @@ void sceneHandler27_throwBat() {
 
 	g_fp->_aniMan->startAnim(MV_MAN27_THROWBET, 0, -1);
 
-	getCurrSceneSc2MotionController()->clearEnabled();
+	getCurrSceneSc2MotionController()->deactivate();
 	getGameLoaderInteractionController()->disableFlag24();
 
 	g_fp->_behaviorManager->setFlagByStaticAniObject(g_fp->_aniMan, 0);
@@ -247,7 +202,7 @@ void sceneHandler27_clickBat(ExCommand *cmd) {
 
 	if (ABS(bx - g_fp->_aniMan->_ox) > 1 || ABS(by - g_fp->_aniMan->_oy) > 1
 		|| g_fp->_aniMan->_movement || g_fp->_aniMan->_statics->_staticsId != ST_MAN_RIGHT) {
-		MessageQueue *mq = getCurrSceneSc2MotionController()->method34(g_fp->_aniMan, bx, by, 1, ST_MAN_RIGHT);
+		MessageQueue *mq = getCurrSceneSc2MotionController()->startMove(g_fp->_aniMan, bx, by, 1, ST_MAN_RIGHT);
 
 		if (mq) {
 			mq->addExCommandToEnd(cmd->createClone());
@@ -268,17 +223,19 @@ void sceneHandler27_maidSwab() {
 }
 
 void sceneHandler27_startBat(StaticANIObject *bat) {
+	debugC(2, kDebugSceneLogic, "scene27: startBat");
+
 	Bat *newbat = new Bat;
 
 	newbat->power = g_vars->scene27_launchPhase * 2.5 + 8.0;
-	newbat->field_10 = 0;
+	newbat->angle = 0;
 	newbat->ani = bat;
 	newbat->powerCos = newbat->power * cos(0.0);
 	newbat->powerSin = newbat->power * sin(0.0);
 	newbat->currX = newbat->powerCos + (double)g_fp->_aniMan->_ox + 42.0;
 	newbat->currY = newbat->powerSin + (double)g_fp->_aniMan->_oy + 58.0;
 
-	bat->_statics = (Statics *)bat->_staticsList[0];
+	bat->_statics = bat->_staticsList[0];
 	bat->setOXY((int)newbat->currX, (int)newbat->currY);
 	bat->_flags |= 4;
 
@@ -291,16 +248,16 @@ void sceneHandler27_startAiming() {
 	g_vars->scene27_dudeIsAiming = false;
 	g_vars->scene27_maxPhaseReached = false;
 
-	g_fp->_aniMan->_callback2 = 0;
+	g_fp->_aniMan->_callback2 = 0; // Really NULL
 
 	g_vars->scene27_launchPhase = g_fp->_aniMan->_movement->_currDynamicPhaseIndex - 6;
 
 	int phase = 21 - g_vars->scene27_launchPhase;
 
-    if (phase < 14)
+	if (phase < 14)
 		phase = 14;
 
-    if (phase > 20)
+	if (phase > 20)
 		phase = 20;
 
 	g_fp->playSound(SND_27_044, 0);
@@ -329,18 +286,20 @@ void sceneHandler27_aimDude() {
 }
 
 void sceneHandler27_wipeDo() {
+	debugC(2, kDebugSceneLogic, "scene27: wipeDo");
+
 	for (uint i = 0; i < g_vars->scene27_bats.size(); i++) {
 		if (g_vars->scene27_bats[i]->currX < 800.0) {
-			g_vars->scene27_bats[i]->field_10 = atan2(800.0 - g_vars->scene27_bats[i]->currX, 520.0 - g_vars->scene27_bats[i]->currY);
+			g_vars->scene27_bats[i]->angle = atan2(520.0 - g_vars->scene27_bats[i]->currY, 800.0 - g_vars->scene27_bats[i]->currX);
 			g_vars->scene27_bats[i]->power += 1.0;
 		}
-    }
+	}
 }
 
 bool sceneHandler27_batFallLogic(uint batn) {
 	Bat *bat = g_vars->scene27_bats[batn];
 
-	int y = (bat->currY - 458.0) * 0.4848484848484849 + 734.0;
+	int y = (int)((bat->currY - 458.0) * 0.4848484848484849 + 734.0);
 
 	if (y >= bat->currX)
 		return false;
@@ -350,17 +309,17 @@ bool sceneHandler27_batFallLogic(uint batn) {
 
 		g_vars->scene27_var07.push_back(bat);
 
-		if (batn != g_vars->scene27_var07.size() - 1)
-			g_vars->scene27_var07.remove_at(batn);
-    } else if (!bat->ani->_movement) {
+		if (batn != g_vars->scene27_bats.size() - 1)
+			g_vars->scene27_bats.remove_at(batn);
+	} else if (!bat->ani->_movement) {
 		bat->ani->startAnim(MV_BTA_FALL, 0, -1);
-    }
+	}
 
 	return true;
 }
 
 bool sceneHandler27_batCalcDistance(int bat1, int bat2) {
-	double at = atan2(g_vars->scene27_bats[bat1]->currX - g_vars->scene27_bats[bat2]->currX, g_vars->scene27_bats[bat1]->currY - g_vars->scene27_bats[bat2]->currY);
+	double at = atan2(g_vars->scene27_bats[bat1]->currY - g_vars->scene27_bats[bat2]->currY, g_vars->scene27_bats[bat1]->currX - g_vars->scene27_bats[bat2]->currX);
 	double dy = g_vars->scene27_bats[bat1]->currY - g_vars->scene27_bats[bat2]->currY;
 	double dx = g_vars->scene27_bats[bat1]->currX - g_vars->scene27_bats[bat2]->currX;
 	double ay = cos(at);
@@ -373,50 +332,73 @@ void sceneHandler27_knockBats(int bat1n, int bat2n) {
 	Bat *bat1 = g_vars->scene27_bats[bat1n];
 	Bat *bat2 = g_vars->scene27_bats[bat2n];
 
-	if (0.0 != bat1->power) {
-		double rndF = (double)g_fp->_rnd->getRandomNumber(32767) * 0.0000009155552842799158 - 0.015
-			+ atan2(bat2->currX - bat1->currX, bat2->currY - bat1->currY);
-		double rndCos = cos(rndF);
-		double rndSin = sin(rndF);
+	debugC(2, kDebugSceneLogic, "scene27: knockBats(%d, %d)", bat1n, bat2n);
 
-		double pow1x = cos(bat1->field_10 - rndF) * (double)((int)(bat2->currX - bat1->currX) >= 0 ? 1 : -1) * bat1->power;
-		double pow1y = sin(bat1->field_10 - rndF) * (double)((int)(bat2->currY - bat1->currY) >= 0 ? 1 : -1) * bat1->power;
+	if (bat1->power != 0.0) {
+		double rndF = (double)g_fp->_rnd.getRandomNumber(32767) * 0.03 / 32767.0 - 0.015
+			+ atan2(bat2->currY - bat1->currY, bat2->currX - bat1->currX);
+
+		double pow1x = cos(bat1->angle - rndF) * ((bat2->currX - bat1->currX) >= 0 ? bat1->power : -bat1->power);
+		double pow1y = sin(bat1->angle - rndF) * ((bat2->currY - bat1->currY) >= 0 ? bat1->power : -bat1->power);
+
+		debugC(3, kDebugSceneLogic, "scene27: knockBats: bat1 from: powerCos: %f powerSin: %f, power: %f, angle: %f",
+				bat1->powerCos, bat1->powerSin, bat1->power, bat1->angle);
 
 		bat1->powerCos -= pow1x * 1.1;
 		bat1->powerSin -= pow1y * 1.1;
 
-		rndF = ((double)g_fp->_rnd->getRandomNumber(32767) * 0.0000009155552842799158 - 0.015
-							   + atan2(bat1->currX - bat2->currX, bat1->currY - bat2->currY));
-		double pow2x = cos(bat2->field_10 - rndF) * (double)((int)(bat1->currX - bat2->currX) >= 0 ? 1 : -1) * bat2->power;
-		double pow2y = sin(bat2->field_10 - rndF) * (double)((int)(bat1->currY - bat2->currY) >= 0 ? 1 : -1) * bat2->power;
+		debugC(3, kDebugSceneLogic, "scene27: knockBats: bat1 to: powerCos: %f powerSin: %f", bat1->powerCos, bat1->powerSin);
+
+		double rndF2 = (double)g_fp->_rnd.getRandomNumber(32767) * 0.03 / 32767.0 - 0.015
+								+ atan2(bat1->currY - bat2->currY, bat1->currX - bat2->currX);
+		double pow2x = cos(bat2->angle - rndF2) * ((bat1->currX - bat2->currX) >= 0 ? bat2->power : -bat2->power);
+		double pow2y = sin(bat2->angle - rndF2) * ((bat1->currY - bat2->currY) >= 0 ? bat2->power : -bat2->power);
+
+		debugC(3, kDebugSceneLogic, "scene27: knockBats: bat2 from: powerCos: %f powerSin: %f, power: %f, angle: %f",
+				bat2->powerCos, bat2->powerSin, bat2->power, bat2->angle);
 
 		bat2->powerCos -= pow2x * 1.1;
 		bat2->powerSin -= pow2y * 1.1;
 
+		debugC(3, kDebugSceneLogic, "scene27: knockBats: bat2 to: powerCos: %f powerSin: %f", bat2->powerCos, bat2->powerSin);
+
+		double rndCos = cos(rndF);
+		double rndSin = sin(rndF);
+
 		double dy = bat1->currY - bat2->currY;
-	    double dx = bat1->currX - bat2->currX;
-		double dist = (sqrt(rndSin * rndSin * 0.25 + rndCos * rndCos) * 54.0 - sqrt(dx * dx + dy * dy)) / cos(rndF - bat1->field_10);
-		bat1->currX -= cos(bat1->field_10) * (dist + 1.0);
-		bat1->currY -= sin(bat1->field_10) * (dist + 1.0);
+		double dx = bat1->currX - bat2->currX;
+		double dist = (sqrt(rndSin * rndSin * 0.25 + rndCos * rndCos) * 54.0 - sqrt(dx * dx + dy * dy)) / cos(rndF - bat1->angle);
+		bat1->currX = bat1->currX - cos(bat1->angle) * (dist + 1.0);
+		bat1->currY = bat1->currY - sin(bat1->angle) * (dist + 1.0);
+
 		bat1->powerCos += pow2x * 0.64;
 
+		debugC(3, kDebugSceneLogic, "scene27: knockBats: bat1 x: %g y: %g", bat1->currX, bat1->currY);
+
 		if (bat1->currX <= 500.0)
 			bat1->powerSin = 0.0;
 		else
 			bat1->powerSin += pow2y * 0.64;
 
-		bat1->field_10 = atan2(bat1->powerCos, bat1->powerSin);
+		bat1->angle = atan2(bat1->powerSin, bat1->powerCos);
 		bat1->power = sqrt(bat1->powerCos * bat1->powerCos + bat1->powerSin * bat1->powerSin);
+
+		debugC(3, kDebugSceneLogic, "scene27: knockBats: bat1 corrected: powerCos: %f powerSin: %f, power: %f, angle: %f",
+				bat1->powerCos, bat1->powerSin, bat1->power, bat1->angle);
+
 		bat2->powerCos += pow1x * 0.64;
 
 		if (bat2->currX <= 500.0)
-			bat2->powerSin = 0;
+			bat2->powerSin = 0.0;
 		else
 			bat2->powerSin += pow1y * 0.64;
 
-		bat2->field_10 = atan2(bat2->powerCos, bat2->powerSin);
+		bat2->angle = atan2(bat2->powerSin, bat2->powerCos);
 		bat2->power = sqrt(bat2->powerCos * bat2->powerCos + bat2->powerSin * bat2->powerSin);
 
+		debugC(3, kDebugSceneLogic, "scene27: knockBats: bat2 corrected: powerCos: %f powerSin: %f, power: %f, angle: %f",
+				bat2->powerCos, bat2->powerSin, bat2->power, bat2->angle);
+
 		g_fp->playSound(SND_27_026, 0);
 	}
 }
@@ -438,6 +420,8 @@ void sceneHandler27_batSetColors(int batn) {
 }
 
 void sceneHandler27_driverPushButton() {
+	debugC(2, kDebugSceneLogic, "scene27: driverPushButton");
+
 	if (g_fp->getObjectState(sO_Driver) == g_fp->getObjectEnumState(sO_Driver, sO_WithSteering)) {
 		g_vars->scene27_driver->changeStatics2(ST_DRV_VENT);
 		chainQueue(QU_DRV_PUSHBUTTON, 1);
@@ -463,17 +447,10 @@ void sceneHandler27_maidSwitchback() {
 }
 
 void sceneHandler27_batLogic() {
-	if (g_vars->scene27_balls.numBalls) {
-		g_vars->scene27_bat = g_vars->scene27_balls.pHead->ani;
+	if (g_vars->scene27_balls.size()) {
+		g_vars->scene27_bat = g_vars->scene27_balls[0];
 
-		g_vars->scene27_balls.pHead = g_vars->scene27_balls.pHead->p0;
-
-		if (g_vars->scene27_balls.pHead)
-			g_vars->scene27_balls.pHead->p0->p1 = 0;
-		else
-			g_vars->scene27_balls.field_8 = 0;
-
-		g_vars->scene27_balls.init(&g_vars->scene27_balls.pHead->p0);
+		g_vars->scene27_balls.remove_at(0);
 
 		int mv;
 
@@ -501,7 +478,7 @@ void sceneHandler27_batLogic() {
 		default:
 			chainQueue(QU_SC27_RESTARTBETS, 1);
 
-			getCurrSceneSc2MotionController()->setEnabled();
+			getCurrSceneSc2MotionController()->activate();
 			getGameLoaderInteractionController()->enableFlag24();
 
 			g_fp->_behaviorManager->setFlagByStaticAniObject(g_fp->_aniMan, 1);
@@ -545,7 +522,7 @@ void sceneHandler27_calcWinArcade() {
 			}
 		}
 
-		if (!g_vars->scene27_balls.numBalls) {
+		if (!g_vars->scene27_balls.size()) {
 			sceneHandler27_driverPushButton();
 			sceneHandler27_maidSwitchback();
 			return;
@@ -558,49 +535,16 @@ void sceneHandler27_calcWinArcade() {
 }
 
 void sceneHandler27_regenBats() {
+	debugC(2, kDebugSceneLogic, "scene27: regenBats");
+
 	g_vars->scene27_wipeIsNeeded = false;
 
 	for (uint i = 0; i < g_vars->scene27_var07.size(); i++) {
 		g_vars->scene27_var07[i]->ani->hide();
 
-		Ball *runPtr = g_vars->scene27_balls.pTail;
-		Ball *lastP = g_vars->scene27_balls.field_8;
 		StaticANIObject *newbat = g_vars->scene27_var07[i]->ani;
 
-		if (!g_vars->scene27_balls.pTail) {
-			g_vars->scene27_balls.cPlex = (byte *)calloc(g_vars->scene27_balls.cPlexLen, sizeof(Ball));
-
-			byte *p1 = g_vars->scene27_balls.cPlex + (g_vars->scene27_balls.cPlexLen - 1) * sizeof(Ball);
-
-			if (g_vars->scene27_balls.cPlexLen - 1 < 0) {
-				runPtr = g_vars->scene27_balls.pTail;
-			} else {
-				runPtr = g_vars->scene27_balls.pTail;
-
-				for (int j = 0; j < g_vars->scene27_balls.cPlexLen; j++) {
-					((Ball *)p1)->p1 = runPtr;
-					runPtr = (Ball *)p1;
-
-					p1 -= sizeof(Ball);
-				}
-
-				g_vars->scene27_balls.pTail = runPtr;
-			}
-		}
-
-		g_vars->scene27_balls.pTail = runPtr->p0;
-		runPtr->p1 = lastP;
-		runPtr->p0 = 0;
-		runPtr->ani = newbat;
-
-		g_vars->scene27_balls.numBalls++;
-
-		if (g_vars->scene27_balls.field_8)
-			g_vars->scene27_balls.field_8->p0 = runPtr;
-		else
-			g_vars->scene27_balls.pHead = runPtr;
-
-		g_vars->scene27_balls.field_8 = runPtr;
+		g_vars->scene27_balls.push_back(newbat);
 	}
 
 	g_vars->scene27_var07.clear();
@@ -619,22 +563,22 @@ void sceneHandler27_animateBats() {
 	for (uint i = 0; i < g_vars->scene27_bats.size(); i++) {
 		Bat *bat = g_vars->scene27_bats[i];
 
-		bat->currX = cos(bat->field_10) * bat->power + bat->currX;
-		bat->currY = sin(bat->field_10) * bat->power + bat->currY;
+		bat->currX = cos(bat->angle) * bat->power + bat->currX;
+		bat->currY = sin(bat->angle) * bat->power + bat->currY;
 
 		bat->ani->setOXY((int)bat->currX, (int)bat->currY);
 		bat->ani->_priority = (int)(600.0 - bat->currY);
 
 		double powerDelta;
 
-		if (cos(bat->field_10) >= 0.0 || bat->currX >= 362.0)
+		if (cos(bat->angle) >= 0.0 || bat->currX >= 362.0)
 			powerDelta = bat->power * 0.035;
 		else
 			powerDelta = bat->power * 0.4;
 
 		bat->power -= powerDelta;
-		bat->powerCos = cos(bat->field_10) * bat->power;
-		bat->powerSin = sin(bat->field_10) * bat->power;
+		bat->powerCos = cos(bat->angle) * bat->power;
+		bat->powerSin = sin(bat->angle) * bat->power;
 
 		if (bat->power >= 0.5)
 			g_vars->scene27_knockCount++;
@@ -695,6 +639,8 @@ int sceneHandler27(ExCommand *cmd) {
 		break;
 
 	case MSG_SC27_STARTWIPE:
+		debugC(2, kDebugSceneLogic, "scene27: STARTWIPE");
+
 		g_vars->scene27_wipeIsNeeded = true;
 
 		g_fp->playSound(SND_27_027, 0);
