@@ -8,12 +8,12 @@
  * modify it under the terms of the GNU General Public License
  * as published by the Free Software Foundation; either version 2
  * of the License, or (at your option) any later version.
-
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
-
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
@@ -28,8 +28,8 @@
 #include "common/system.h"
 #include "common/textconsole.h"
 #include "engines/util.h"
-#include "graphics/decoders/jpeg.h"
-#include "graphics/decoders/pict.h"
+#include "graphics/palette.h"
+#include "image/pict.h"
 
 namespace Mohawk {
 
@@ -38,23 +38,23 @@ MystGraphics::MystGraphics(MohawkEngine_Myst* vm) : GraphicsManager(), _vm(vm) {
 
 	_viewport = Common::Rect(544, 332);
 
-	// The original version of Myst could run in 8bpp color too.
-	// However, it dithered videos to 8bpp and they looked considerably
-	// worse (than they already did :P). So we're not even going to
-	// support 8bpp mode in Myst (Myst ME required >8bpp anyway).
-	initGraphics(_viewport.width(), _viewport.height(), true, NULL); // What an odd screen size!
+	if (_vm->getFeatures() & GF_ME) {
+		// High color
+		initGraphics(_viewport.width(), _viewport.height(), nullptr);
+
+		if (_vm->_system->getScreenFormat().bytesPerPixel == 1)
+			error("Myst ME requires greater than 256 colors to run");
+	} else {
+		// Paletted
+		initGraphics(_viewport.width(), _viewport.height());
+		clearScreenPalette();
+	}
 
 	_pixelFormat = _vm->_system->getScreenFormat();
 
-	if (_pixelFormat.bytesPerPixel == 1)
-		error("Myst requires greater than 256 colors to run");
-
 	// Initialize our buffer
 	_backBuffer = new Graphics::Surface();
 	_backBuffer->create(_vm->_system->getWidth(), _vm->_system->getHeight(), _pixelFormat);
-
-	_nextAllowedDrawTime = _vm->_system->getMillis();
-	_enableDrawingTimeSimulation = 0;
 }
 
 MystGraphics::~MystGraphics() {
@@ -69,7 +69,7 @@ MohawkSurface *MystGraphics::decodeImage(uint16 id) {
 	// if it's a PICT or WDIB resource. If it's Myst ME it's most likely a PICT, and if it's
 	// original it's definitely a WDIB. However, Myst ME throws us another curve ball in
 	// that PICT resources can contain WDIB's instead of PICT's.
-	Common::SeekableReadStream *dataStream = NULL;
+	Common::SeekableReadStream *dataStream = nullptr;
 
 	if (_vm->getFeatures() & GF_ME && _vm->hasResource(ID_PICT, id)) {
 		// The PICT resource exists. However, it could still contain a MystBitmap
@@ -82,7 +82,7 @@ MohawkSurface *MystGraphics::decodeImage(uint16 id) {
 
 	bool isPict = false;
 
-	if (_vm->getFeatures() & GF_ME) {
+	if ((_vm->getFeatures() & GF_ME) && dataStream->size() > 512 + 10 + 4) {
 		// Here we detect whether it's really a PICT or a WDIB. Since a MystBitmap
 		// would be compressed, there's no way to detect for the BM without a hack.
 		// So, we search for the PICT version opcode for detection.
@@ -91,24 +91,102 @@ MohawkSurface *MystGraphics::decodeImage(uint16 id) {
 		dataStream->seek(0);
 	}
 
-	MohawkSurface *mhkSurface = 0;
+	MohawkSurface *mhkSurface = nullptr;
 
 	if (isPict) {
-		Graphics::PICTDecoder pict;
+		Image::PICTDecoder pict;
 
 		if (!pict.loadStream(*dataStream))
-			error("Could not decode Myst ME PICT");
+			error("Could not decode Myst ME PICT %d", id);
+
+		delete dataStream;
 
 		mhkSurface = new MohawkSurface(pict.getSurface()->convertTo(_pixelFormat));
 	} else {
 		mhkSurface = _bmpDecoder->decodeImage(dataStream);
-		mhkSurface->convertToTrueColor();
+
+		if (_vm->getFeatures() & GF_ME) {
+			mhkSurface->convertToTrueColor();
+		} else {
+			remapSurfaceToSystemPalette(mhkSurface);
+		}
 	}
 
 	assert(mhkSurface);
+	applyImagePatches(id, mhkSurface);
 	return mhkSurface;
 }
 
+void MystGraphics::applyImagePatches(uint16 id, const MohawkSurface *mhkSurface) const {
+
+	// In the English ME version of the game, the instructions found on Stoneship
+	// to open the vault are incorrect. They are:
+	//     Turn every one of [these switches to the] "off" position.
+	// They should be:
+	//     Turn every one of [these switches to the] "on" position.
+	//
+	// Here we stomp over the "off" with an "on".
+	// The fixed image was provided by dafioram in bug Trac#10115.
+	if (id == 2019 && _vm->getFeatures() & GF_ME && _vm->getLanguage() == Common::EN_ANY) {
+		static const byte markerSwitchInstructionsFixPic[] = {
+				0x1d, 0x1c, 0x19, 0x19, 0x19, 0x19, 0x1c, 0x19, 0x19, 0x17, 0x19, 0x19, 0x19, 0x19, 0x19,
+				0x1e, 0x1e, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19,
+				0x1c, 0x19, 0x19, 0x19, 0x1c, 0x19, 0x19, 0x19, 0x1c, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19,
+				0x1d, 0x1e, 0x16, 0x0d, 0x0e, 0x12, 0x19, 0x19, 0x17, 0x10, 0x06, 0x05, 0x19, 0x19, 0x19,
+				0x1e, 0x1e, 0x10, 0x13, 0x1c, 0x11, 0x0d, 0x19, 0x12, 0x09, 0x16, 0x04, 0x18, 0x18, 0x19,
+				0x1e, 0x1a, 0x03, 0x1b, 0x1c, 0x17, 0x02, 0x15, 0x13, 0x00, 0x19, 0x06, 0x18, 0x19, 0x18,
+				0x1e, 0x1e, 0x01, 0x1b, 0x1c, 0x1b, 0x02, 0x15, 0x13, 0x00, 0x19, 0x07, 0x0a, 0x19, 0x18,
+				0x1e, 0x1c, 0x0c, 0x0e, 0x14, 0x0c, 0x0c, 0x19, 0x0b, 0x00, 0x19, 0x00, 0x08, 0x19, 0x19,
+				0x1e, 0x1c, 0x19, 0x14, 0x0f, 0x0f, 0x14, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x17,
+				0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19, 0x19,
+				0x1c, 0x1c, 0x1e, 0x19, 0x19, 0x19, 0x17, 0x19, 0x19, 0x17, 0x19, 0x19, 0x19, 0x19, 0x19
+		};
+
+		static const byte markerSwitchInstructionsFixPal[] = {
+				0x00, 0x00, 0x00,
+				0x10, 0x08, 0x08,
+				0x18, 0x10, 0x10,
+				0x28, 0x10, 0x08,
+				0x20, 0x18, 0x18,
+				0x28, 0x20, 0x20,
+				0x38, 0x20, 0x10,
+				0x30, 0x28, 0x20,
+				0x38, 0x30, 0x28,
+				0x40, 0x38, 0x28,
+				0x48, 0x38, 0x28,
+				0x48, 0x40, 0x30,
+				0x50, 0x48, 0x38,
+				0x50, 0x48, 0x40,
+				0x60, 0x50, 0x38,
+				0x68, 0x58, 0x40,
+				0x68, 0x58, 0x48,
+				0x70, 0x60, 0x50,
+				0x78, 0x68, 0x50,
+				0x80, 0x70, 0x50,
+				0x80, 0x78, 0x60,
+				0x88, 0x80, 0x60,
+				0x98, 0x90, 0x70,
+				0xb0, 0xa0, 0x78,
+				0xb8, 0xa8, 0x8d,
+				0xb8, 0xa8, 0x90,
+				0xb8, 0xb0, 0x88,
+				0xc0, 0xb8, 0x90,
+				0xd8, 0xcc, 0x98,
+				0xd0, 0xe0, 0xc8,
+				0xf0, 0xe4, 0xc8
+		};
+
+		Graphics::Surface fixSurf;
+		fixSurf.create(15, 11, Graphics::PixelFormat::createFormatCLUT8());
+		fixSurf.copyRectToSurface(markerSwitchInstructionsFixPic, fixSurf.w, 0, 0, fixSurf.w, fixSurf.h);
+		fixSurf.convertToInPlace(_pixelFormat, markerSwitchInstructionsFixPal);
+
+		mhkSurface->getSurface()->copyRectToSurface(fixSurf, 171, 208, Common::Rect(fixSurf.w, fixSurf.h));
+
+		fixSurf.free();
+	}
+}
+
 void MystGraphics::copyImageSectionToScreen(uint16 image, Common::Rect src, Common::Rect dest) {
 	Graphics::Surface *surface = findImage(image)->getSurface();
 
@@ -146,13 +224,12 @@ void MystGraphics::copyImageSectionToScreen(uint16 image, Common::Rect src, Comm
 	debug(3, "\twidth: %d", width);
 	debug(3, "\theight: %d", height);
 
-	simulatePreviousDrawDelay(dest);
-
 	_vm->_system->copyRectToScreen(surface->getBasePtr(src.left, top), surface->pitch, dest.left, dest.top, width, height);
 }
 
 void MystGraphics::copyImageSectionToBackBuffer(uint16 image, Common::Rect src, Common::Rect dest) {
-	Graphics::Surface *surface = findImage(image)->getSurface();
+	MohawkSurface *mhkSurface = findImage(image);
+	Graphics::Surface *surface = mhkSurface->getSurface();
 
 	// Make sure the image is bottom aligned in the dest rect
 	dest.top = dest.bottom - MIN<int>(surface->h, dest.height());
@@ -191,6 +268,13 @@ void MystGraphics::copyImageSectionToBackBuffer(uint16 image, Common::Rect src,
 
 	for (uint16 i = 0; i < height; i++)
 		memcpy(_backBuffer->getBasePtr(dest.left, i + dest.top), surface->getBasePtr(src.left, top + i), width * surface->format.bytesPerPixel);
+
+	if (!(_vm->getFeatures() & GF_ME)) {
+		// Make sure the palette is set
+		assert(mhkSurface->getPalette());
+		memcpy(_palette, mhkSurface->getPalette(), 256 * 3);
+		setPaletteToScreen();
+	}
 }
 
 void MystGraphics::copyImageToScreen(uint16 image, Common::Rect dest) {
@@ -204,17 +288,11 @@ void MystGraphics::copyImageToBackBuffer(uint16 image, Common::Rect dest) {
 void MystGraphics::copyBackBufferToScreen(Common::Rect r) {
 	r.clip(_viewport);
 
-	simulatePreviousDrawDelay(r);
-
 	_vm->_system->copyRectToScreen(_backBuffer->getBasePtr(r.left, r.top), _backBuffer->pitch, r.left, r.top, r.width(), r.height());
 }
 
 void MystGraphics::runTransition(TransitionType type, Common::Rect rect, uint16 steps, uint16 delay) {
 
-	// Do not artificially delay during transitions
-	int oldEnableDrawingTimeSimulation = _enableDrawingTimeSimulation;
-	_enableDrawingTimeSimulation = 0;
-
 	switch (type) {
 	case kTransitionLeftToRight:	{
 			debugC(kDebugView, "Left to Right");
@@ -225,17 +303,14 @@ void MystGraphics::runTransition(TransitionType type, Common::Rect rect, uint16
 				area.left = rect.left + step * i;
 				area.right = area.left + step;
 
-				_vm->_system->delayMillis(delay);
-
 				copyBackBufferToScreen(area);
-				_vm->_system->updateScreen();
+				_vm->wait(delay);
 			}
 			if (area.right < rect.right) {
 				area.left = area.right;
 				area.right = rect.right;
 
 				copyBackBufferToScreen(area);
-				_vm->_system->updateScreen();
 			}
 		}
 		break;
@@ -248,17 +323,14 @@ void MystGraphics::runTransition(TransitionType type, Common::Rect rect, uint16
 				area.right = rect.right - step * i;
 				area.left = area.right - step;
 
-				_vm->_system->delayMillis(delay);
-
 				copyBackBufferToScreen(area);
-				_vm->_system->updateScreen();
+				_vm->wait(delay);
 			}
 			if (area.left > rect.left) {
 				area.right = area.left;
 				area.left = rect.left;
 
 				copyBackBufferToScreen(area);
-				_vm->_system->updateScreen();
 			}
 		}
 		break;
@@ -274,8 +346,8 @@ void MystGraphics::runTransition(TransitionType type, Common::Rect rect, uint16
 			debugC(kDebugView, "Dissolve");
 
 			for (int16 step = 0; step < 8; step++) {
-				simulatePreviousDrawDelay(rect);
 				transitionDissolve(rect, step);
+				_vm->doFrame();
 			}
 		}
 		break;
@@ -288,17 +360,14 @@ void MystGraphics::runTransition(TransitionType type, Common::Rect rect, uint16
 				area.top = rect.top + step * i;
 				area.bottom = area.top + step;
 
-				_vm->_system->delayMillis(delay);
-
 				copyBackBufferToScreen(area);
-				_vm->_system->updateScreen();
+				_vm->wait(delay);
 			}
 			if (area.bottom < rect.bottom) {
 				area.top = area.bottom;
 				area.bottom = rect.bottom;
 
 				copyBackBufferToScreen(area);
-				_vm->_system->updateScreen();
 			}
 		}
 		break;
@@ -311,17 +380,14 @@ void MystGraphics::runTransition(TransitionType type, Common::Rect rect, uint16
 				area.bottom = rect.bottom - step * i;
 				area.top = area.bottom - step;
 
-				_vm->_system->delayMillis(delay);
-
 				copyBackBufferToScreen(area);
-				_vm->_system->updateScreen();
+				_vm->wait(delay);
 			}
 			if (area.top > rect.top) {
 				area.bottom = area.top;
 				area.top = rect.top;
 
 				copyBackBufferToScreen(area);
-				_vm->_system->updateScreen();
 			}
 		}
 		break;
@@ -347,13 +413,10 @@ void MystGraphics::runTransition(TransitionType type, Common::Rect rect, uint16
 		break;
 	case kTransitionCopy:
 		copyBackBufferToScreen(rect);
-		_vm->_system->updateScreen();
 		break;
 	default:
 		error("Unknown transition %d", type);
 	}
-
-	_enableDrawingTimeSimulation = oldEnableDrawingTimeSimulation;
 }
 
 void MystGraphics::transitionDissolve(Common::Rect rect, uint step) {
@@ -420,19 +483,22 @@ void MystGraphics::transitionDissolve(Common::Rect rect, uint step) {
 
 		for (uint16 x = rect.left; x < rect.right; x++) {
 			if (linePattern[x % 4]) {
-				if (_pixelFormat.bytesPerPixel == 2) {
-					uint16 *dst = (uint16 *)screen->getBasePtr(x, y);
-					*dst = *(const uint16 *)_backBuffer->getBasePtr(x, y);
-				} else {
-					uint32 *dst = (uint32 *)screen->getBasePtr(x, y);
-					*dst = *(const uint32 *)_backBuffer->getBasePtr(x, y);
+				switch (_pixelFormat.bytesPerPixel) {
+				case 1:
+					*((byte *)screen->getBasePtr(x, y)) = *((const byte *)_backBuffer->getBasePtr(x, y));
+					break;
+				case 2:
+					*((uint16 *)screen->getBasePtr(x, y)) = *((const uint16 *)_backBuffer->getBasePtr(x, y));
+					break;
+				case 4:
+					*((uint32 *)screen->getBasePtr(x, y)) = *((const uint32 *)_backBuffer->getBasePtr(x, y));
+					break;
 				}
 			}
 		}
 	}
 
 	_vm->_system->unlockScreen();
-	_vm->_system->updateScreen();
 }
 
 void MystGraphics::transitionSlideToLeft(Common::Rect rect, uint16 steps, uint16 delay) {
@@ -446,17 +512,13 @@ void MystGraphics::transitionSlideToLeft(Common::Rect rect, uint16 steps, uint16
 		dstRect.right = dstRect.left + step * stepWidth;
 		srcRect.left = srcRect.right - step * stepWidth;
 
-		_vm->_system->delayMillis(delay);
-
-		simulatePreviousDrawDelay(dstRect);
 		_vm->_system->copyRectToScreen(_backBuffer->getBasePtr(dstRect.left, dstRect.top),
 				_backBuffer->pitch, srcRect.left, srcRect.top, srcRect.width(), srcRect.height());
-		_vm->_system->updateScreen();
+		_vm->wait(delay);
 	}
 
 	if (dstRect.right != rect.right) {
 		copyBackBufferToScreen(rect);
-		_vm->_system->updateScreen();
 	}
 }
 
@@ -471,17 +533,13 @@ void MystGraphics::transitionSlideToRight(Common::Rect rect, uint16 steps, uint1
 		dstRect.left = dstRect.right - step * stepWidth;
 		srcRect.right = srcRect.left + step * stepWidth;
 
-		_vm->_system->delayMillis(delay);
-
-		simulatePreviousDrawDelay(dstRect);
 		_vm->_system->copyRectToScreen(_backBuffer->getBasePtr(dstRect.left, dstRect.top),
 				_backBuffer->pitch, srcRect.left, srcRect.top, srcRect.width(), srcRect.height());
-		_vm->_system->updateScreen();
+		_vm->wait(delay);
 	}
 
 	if (dstRect.left != rect.left) {
 		copyBackBufferToScreen(rect);
-		_vm->_system->updateScreen();
 	}
 }
 
@@ -496,18 +554,14 @@ void MystGraphics::transitionSlideToTop(Common::Rect rect, uint16 steps, uint16
 		dstRect.bottom = dstRect.top + step * stepWidth;
 		srcRect.top = srcRect.bottom - step * stepWidth;
 
-		_vm->_system->delayMillis(delay);
-
-		simulatePreviousDrawDelay(dstRect);
 		_vm->_system->copyRectToScreen(_backBuffer->getBasePtr(dstRect.left, dstRect.top),
 				_backBuffer->pitch, srcRect.left, srcRect.top, srcRect.width(), srcRect.height());
-		_vm->_system->updateScreen();
+		_vm->wait(delay);
 	}
 
 
 	if (dstRect.bottom < rect.bottom) {
 		copyBackBufferToScreen(rect);
-		_vm->_system->updateScreen();
 	}
 }
 
@@ -522,18 +576,14 @@ void MystGraphics::transitionSlideToBottom(Common::Rect rect, uint16 steps, uint
 		dstRect.top = dstRect.bottom - step * stepWidth;
 		srcRect.bottom = srcRect.top + step * stepWidth;
 
-		_vm->_system->delayMillis(delay);
-
-		simulatePreviousDrawDelay(dstRect);
 		_vm->_system->copyRectToScreen(_backBuffer->getBasePtr(dstRect.left, dstRect.top),
 				_backBuffer->pitch, srcRect.left, srcRect.top, srcRect.width(), srcRect.height());
-		_vm->_system->updateScreen();
+		_vm->wait(delay);
 	}
 
 
 	if (dstRect.top > rect.top) {
 		copyBackBufferToScreen(rect);
-		_vm->_system->updateScreen();
 	}
 }
 
@@ -548,14 +598,12 @@ void MystGraphics::transitionPartialToRight(Common::Rect rect, uint32 width, uin
 		dstRect.right = dstRect.left + step * stepWidth;
 		srcRect.left = srcRect.right - step * stepWidth;
 
-		simulatePreviousDrawDelay(dstRect);
 		_vm->_system->copyRectToScreen(_backBuffer->getBasePtr(dstRect.left, dstRect.top),
 				_backBuffer->pitch, srcRect.left, srcRect.top, srcRect.width(), srcRect.height());
-		_vm->_system->updateScreen();
+		_vm->doFrame();
 	}
 
 	copyBackBufferToScreen(rect);
-	_vm->_system->updateScreen();
 }
 
 void MystGraphics::transitionPartialToLeft(Common::Rect rect, uint32 width, uint32 steps) {
@@ -569,14 +617,12 @@ void MystGraphics::transitionPartialToLeft(Common::Rect rect, uint32 width, uint
 		dstRect.left = dstRect.right - step * stepWidth;
 		srcRect.right = srcRect.left + step * stepWidth;
 
-		simulatePreviousDrawDelay(dstRect);
 		_vm->_system->copyRectToScreen(_backBuffer->getBasePtr(dstRect.left, dstRect.top),
 				_backBuffer->pitch, srcRect.left, srcRect.top, srcRect.width(), srcRect.height());
-		_vm->_system->updateScreen();
+		_vm->doFrame();
 	}
 
 	copyBackBufferToScreen(rect);
-	_vm->_system->updateScreen();
 }
 
 void MystGraphics::drawRect(Common::Rect rect, RectState state) {
@@ -589,11 +635,11 @@ void MystGraphics::drawRect(Common::Rect rect, RectState state) {
 	Graphics::Surface *screen = _vm->_system->lockScreen();
 
 	if (state == kRectEnabled)
-		screen->frameRect(rect, _pixelFormat.RGBToColor(0, 255, 0));
+		screen->frameRect(rect, (_vm->getFeatures() & GF_ME) ? _pixelFormat.RGBToColor(0, 255, 0) : 250);
 	else if (state == kRectUnreachable)
-		screen->frameRect(rect, _pixelFormat.RGBToColor(0, 0, 255));
+		screen->frameRect(rect, (_vm->getFeatures() & GF_ME) ? _pixelFormat.RGBToColor(0, 0, 255) : 252);
 	else
-		screen->frameRect(rect, _pixelFormat.RGBToColor(255, 0, 0));
+		screen->frameRect(rect, (_vm->getFeatures() & GF_ME) ? _pixelFormat.RGBToColor(255, 0, 0) : 249);
 
 	_vm->_system->unlockScreen();
 }
@@ -602,78 +648,146 @@ void MystGraphics::drawLine(const Common::Point &p1, const Common::Point &p2, ui
 	_backBuffer->drawLine(p1.x, p1.y, p2.x, p2.y, color);
 }
 
-void MystGraphics::enableDrawingTimeSimulation(bool enable) {
-	if (enable)
-		_enableDrawingTimeSimulation++;
-	else
-		_enableDrawingTimeSimulation--;
-
-	if (_enableDrawingTimeSimulation < 0)
-		_enableDrawingTimeSimulation = 0;
-}
-
-void MystGraphics::simulatePreviousDrawDelay(const Common::Rect &dest) {
-	uint32 time = 0;
-
-	if (_enableDrawingTimeSimulation) {
-		time = _vm->_system->getMillis();
-
-		// Do not draw anything new too quickly after the previous draw call
-		// so that images stay at least a little while on screen
-		// This is enabled only for scripted draw calls
-		if (time < _nextAllowedDrawTime)
-			_vm->_system->delayMillis(_nextAllowedDrawTime - time);
-	}
-
-	// Next draw call allowed at DELAY + AERA * COEFF milliseconds from now
-	time = _vm->_system->getMillis();
-	_nextAllowedDrawTime = time + _constantDrawDelay + dest.height() * dest.width() / _proportionalDrawDelay;
-}
-
-void MystGraphics::copyBackBufferToScreenWithSaturation(int16 saturation) {
-	Graphics::Surface *screen = _vm->_system->lockScreen();
-
-	for (uint16 y = 0; y < _viewport.height(); y++)
-		for (uint16 x = 0; x < _viewport.width(); x++) {
-			uint32 color;
-			uint8 r, g, b;
-
-			if (_pixelFormat.bytesPerPixel == 2)
-				color = *(const uint16 *)_backBuffer->getBasePtr(x, y);
-			else
-				color = *(const uint32 *)_backBuffer->getBasePtr(x, y);
-
-			_pixelFormat.colorToRGB(color, r, g, b);
-
-			r = CLIP<int16>((int16)r - saturation, 0, 255);
-			g = CLIP<int16>((int16)g - saturation, 0, 255);
-			b = CLIP<int16>((int16)b - saturation, 0, 255);
-
-			color = _pixelFormat.RGBToColor(r, g, b);
-
-			if (_pixelFormat.bytesPerPixel == 2) {
-				uint16 *dst = (uint16 *)screen->getBasePtr(x, y);
-				*dst = color;
-			} else {
-				uint32 *dst = (uint32 *)screen->getBasePtr(x, y);
-				*dst = color;
-			}
-		}
-
-	_vm->_system->unlockScreen();
-	_vm->_system->updateScreen();
-}
-
 void MystGraphics::fadeToBlack() {
-	for (int16 i = 0; i < 256; i += 32) {
-		copyBackBufferToScreenWithSaturation(i);
+	// This is only for the demo
+	assert(!(_vm->getFeatures() & GF_ME));
+
+	// Linear fade in 64 steps
+	for (int i = 63; i >= 0; i--) {
+		byte palette[256 * 3];
+		byte *src = _palette;
+		byte *dst = palette;
+
+		for (uint j = 0; j < sizeof(palette); j++)
+			*dst++ = *src++ * i / 64;
+
+		_vm->_system->getPaletteManager()->setPalette(palette, 0, 256);
+		_vm->doFrame();
 	}
 }
 
 void MystGraphics::fadeFromBlack() {
-	for (int16 i = 256; i >= 0; i -= 32) {
-		copyBackBufferToScreenWithSaturation(i);
+	// This is only for the demo
+	assert(!(_vm->getFeatures() & GF_ME));
+
+	copyBackBufferToScreen(_viewport);
+
+	// Linear fade in 64 steps
+	for (int i = 0; i < 64; i++) {
+		byte palette[256 * 3];
+		byte *src = _palette;
+		byte *dst = palette;
+
+		for (uint j = 0; j < sizeof(palette); j++)
+			*dst++ = *src++ * i / 64;
+
+		_vm->_system->getPaletteManager()->setPalette(palette, 0, 256);
+		_vm->doFrame();
+	}
+
+	// Set the full palette
+	_vm->_system->getPaletteManager()->setPalette(_palette, 0, 256);
+}
+
+void MystGraphics::clearScreenPalette() {
+	// Set the palette to all black
+	byte palette[256 * 3];
+	memset(palette, 0, sizeof(palette));
+	_vm->_system->getPaletteManager()->setPalette(palette, 0, 256);
+}
+
+void MystGraphics::remapSurfaceToSystemPalette(MohawkSurface *mhkSurface) {
+	// Entries [0, 9] of the palette
+	static const byte lowPalette[] = {
+		0x00, 0x00, 0x00,
+		0x80, 0x00, 0x00,
+		0x00, 0x80, 0x00,
+		0x80, 0x80, 0x00,
+		0x00, 0x00, 0x80,
+		0x80, 0x00, 0x80,
+		0x00, 0x80, 0x80,
+		0xC0, 0xC0, 0xC0,
+		0xC0, 0xDC, 0xC0,
+		0xA6, 0xCA, 0xF0
+	};
+
+	// Entries [246, 255] of the palette
+	static const byte highPalette[] = {
+		0xFF, 0xFB, 0xF0,
+		0xA0, 0xA0, 0xA4,
+		0x80, 0x80, 0x80,
+		0xFF, 0x00, 0x00,
+		0x00, 0xFF, 0x00,
+		0xFF, 0xFF, 0x00,
+		0x00, 0x00, 0xFF,
+		0xFF, 0x00, 0xFF,
+		0x00, 0xFF, 0xFF,
+		0xFF, 0xFF, 0xFF
+	};
+
+	byte *originalPalette = mhkSurface->getPalette();
+
+	// The target palette is made of the Windows reserved palette, and colors 10 to 245
+	// of the bitmap palette. Entries 0 to 9 and 246 to 255 of the bitmap palette are
+	// discarded.
+	byte targetPalette[256 * 3];
+	memcpy(targetPalette, lowPalette, sizeof(lowPalette));
+	memcpy(targetPalette + sizeof(lowPalette), originalPalette + sizeof(lowPalette), sizeof(_palette) - sizeof(lowPalette) - sizeof(highPalette));
+	memcpy(targetPalette + sizeof(_palette) - sizeof(highPalette), highPalette, sizeof(highPalette));
+
+	// Remap the discarded entries from the bitmap palette using the target palette.
+	byte lowColorMap[ARRAYSIZE(lowPalette) / 3];
+	byte highColorMap[ARRAYSIZE(highPalette) / 3];
+
+	for (uint i = 0; i < ARRAYSIZE(lowColorMap); i++) {
+		uint colorIndex = 3 * i;
+		byte red = originalPalette[colorIndex + 0];
+		byte green = originalPalette[colorIndex + 1];
+		byte blue = originalPalette[colorIndex + 2];
+
+		lowColorMap[i] = getColorIndex(targetPalette, red, green, blue);
+	}
+
+	for (uint i = 0; i < ARRAYSIZE(highColorMap); i++) {
+		uint colorIndex = 3 * (i + 246);
+		byte red = originalPalette[colorIndex + 0];
+		byte green = originalPalette[colorIndex + 1];
+		byte blue = originalPalette[colorIndex + 2];
+
+		highColorMap[i] = getColorIndex(targetPalette, red, green, blue);
+	}
+
+	// Replace the original palette with the target palette
+	memcpy(originalPalette, targetPalette, sizeof(targetPalette));
+
+	// Remap the pixel data to the target palette
+	Graphics::Surface *surface = mhkSurface->getSurface();
+	byte *pixels = (byte *) surface->getPixels();
+
+	for (int i = 0; i < surface->w * surface->h; i++) {
+		if (pixels[i] < ARRAYSIZE(lowColorMap)) {
+			pixels[i] = lowColorMap[pixels[i]];
+		} else if (pixels[i] >= 246) {
+			pixels[i] = highColorMap[pixels[i] - 246];
+		}
 	}
 }
 
+byte MystGraphics::getColorIndex(const byte *palette, byte red, byte green, byte blue) {
+	for (uint i = 0; i < 256; i++) {
+		if (palette[(3 * i) + 0] == red && palette[(3 * i) + 1] == green && palette[(3 * i) + 2] == blue) {
+			return i;
+		}
+	}
+
+	// GDI actually chooses the nearest color if no exact match is found,
+	// but this should not happen in Myst
+	debug(1, "Color (%d, %d, %d) not in target palette", red, green, blue);
+	return 0;
+}
+
+void MystGraphics::setPaletteToScreen() {
+	_vm->_system->getPaletteManager()->setPalette(_palette, 0, 256);
+}
+
 } // End of namespace Mohawk
